#!/usr/bin/perl

# bindery2ldif - Convert a Netware Bindery to LDIF
# Copyright (C) 2003 Arthur de Jong
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This perl script parses the output from the nwbols, nwboprops and nwbpvalues
# commands available in the ncpfs package which can be downloaded from
# ftp://platan.vc.cvut.cz/pub/linux/ncpfs/

# This script was written for the one-time import of Netware 3.11 Bindery data
# into a LDAP server. It is not really maintained or supported but it is
# distributed in the hope that is is marginally useful.
#
# The ldif data is printed on stdout so you should redirect it to a file. Some
# warnings may be printed to stderr (e.g. nwbpvalues: Could not find property)
# you can almost always safely ignore those. The LDIF data contains the bindery
# data as comments.
#
# All the settings are coded into the script. Marginal support for merging
# it into an existing configuration is available. uid's are looked up on the
# current system, modify the get_uid() and get_gid() functions if you don't
# like this. For most purposes it should be sufficient to edit the top of
# this file.

# base DN for domain
my $basedn='dc=example,dc=com';

# base DN for all users (should be ou)
my $user_basedn='ou=people,'.$basedn;

# base DN for all groups (should be ou)
my $group_basedn='ou=groups,'.$basedn;

# all users will get this as their primary group
my $user_group='100';

# starting userid that is used when handing out uids
# uid's will be looked up when possible first in the current passwd file
my $user_id_startat=2000;

# starting gid that are handed out to all the groups
my $group_id_startat=101;

# base directory for home directories (/username is appended to it)
my $homebase="/home";

# login shell that should be set
my $loginshell="/bin/bash";

# location of the home directory for smb accounts
my $smbhome='\\fs\home';

# value of the homeDrive attribute that all the users will get
my $homedrive='H';

# parameters that the nw* command should get
# you could add -P password
my $nwcmdargs='-U SUPERVISOR';

# users that should not be exported
my @skipusers=( 'guest', 'supervisor', 'backup' );

# groups that should not be exported
my @skipgroups=( 'everyone' );

# flag to indicate if bindery information should be dumped as LDIF comments
my $dumpbindery=1;

# dump domain record, just in case
print "dn: ".$basedn."\n";
print "objectClass: dcObject\n";
$basedn=~/^dc=([^,]*),.*$/;
print "dc: ".$1."\n";
print "\n";

# dump user base DN record, just in case
print "dn: ".$user_basedn."\n";
print "objectClass: organizationalUnit\n";
$user_basedn=~/^ou=([^,]*),.*$/;
print "ou: ".$1."\n";
print "\n";

# dump group base DN record, just in case
print "dn: ".$group_basedn."\n";
print "objectClass: organizationalUnit\n";
$group_basedn=~/^ou=([^,]*),.*$/;
print "ou: ".$1."\n";
print "\n";

# get all objects from the bindery
open NWBOLS, "nwbols $nwcmdargs -v|";
while ( <NWBOLS> )
{
  chomp;
  my ($object_name,$object_id,$object_type,$object_flags,$object_secbyte,$object_props) = split(/ /);
  # dump the bindery information into LDIF comments
  if ($dumpbindery)
  {
    print "################################################################################\n";
    if ( $object_type eq '0001' )
    {
      print "# USER $object_name (id=$object_id)\n";
      print "# PROPERTIES ".join(" ",get_properties($object_name,'0001'))."\n";
    }
    elsif ( $object_type eq '0002' )
    {
      print "# GROUP $object_name (id=$object_id)\n";
    }
    elsif ( $object_type eq '0004' or $object_type = '0107' )
    {
      print "# SERVER $object_name (id=$object_id)\n"
    }
    else
    {
      print "# UNKNOWN $object_name (id=$object_id)\n"
    }
    open NWBOPROPS,"nwboprops $nwcmdargs -o $object_name -t $object_type -v|";
    while ( <NWBOPROPS> )
    {
      chomp;
      ($property_name,$property_flags,$property_secbyte,$property_valflag) = split(/ /);
      print "#   PROPERTY $property_name\n";
      open NWBPVALUES,"nwbpvalues $nwcmdargs -o $object_name -t $object_type -p \"$property_name\" -v|";
      while ( <NWBPVALUES> )
      {
        chomp;
        if ( /^(.*) ([0-9A-F]{8}) ([0-9]{4})$/ )
        {
          print "#     VALUE $1 (id=$2, type=$3)\n";
        }
        else
        {
          print "#     VALUE $_\n";
        }
      }
      close NWBPVALUES;
    }
    close NWBOPROPS;
  }
  # dump binery element (in case of user or group)
  if ( $object_type eq '0001' )
  {
    dump_user($object_name,$object_id,$object_flags,$object_secbyte,$object_props) 
      if ! grep {lc($_) eq lc($object_name)} @skipusers;
  }
  elsif ( $object_type eq '0002' )
  {
    dump_group($object_name,$object_id,$object_flags,$object_secbyte,$object_props)
      if ! grep {lc($_) eq lc($object_name)} @skipgroups;
  }
  print "\n";
}
close NWBOLS;

sub get_properties
{
  my ($object_name,$object_type) = @_;
  my @properties;
  open NWBOPROPS,"nwboprops $nwcmdargs -o $object_name -t $object_type -v|";
  while ( <NWBOPROPS> )
  {
    chomp;
    my ($property_name,$property_flags,$property_secbyte,$property_valflag) = split(/ /);
    push @properties,$property_name;
  }
  close NWBOPROPS;
  return @properties;
}

sub get_property_value
{
  my ($object_name,$object_type,$property_name) = @_;
  my @values;
  open NWBPVALUES,"nwbpvalues $nwcmdargs -o $object_name -t $object_type -p \"$property_name\" -v|";
  while ( <NWBPVALUES> )
  {
    chomp;
    push @values,$_;
  }
  close NWBPVALUES;
  return @values;
}

sub dump_user
{
  my ($user_name,$user_id,$user_flags,$user_secbyte,$user_props) = @_;
  my $uid=get_uid(lc($user_name));
  print "dn: uid=".lc($user_name).",".$user_basedn."\n";
  print "objectClass: top\n";
  print "objectClass: account\n";
  print "objectClass: posixAccount\n";
  print "objectClass: shadowAccount\n";
  print "objectClass: person\n";
  print "objectClass: organizationalPerson\n";
  print "objectClass: inetOrgPerson\n";
  print "objectClass: trustAccount\n";
  print "objectClass: sambaAccount\n";
  print "uid: ".lc($user_name)."\n";
  print "uidNumber: ".$uid."\n";
  print "gidNumber: 100\n";
  my ($username,@dummy)=ansi2utf(get_property_value($user_name,'0001','IDENTIFICATION'));
  if ( $username )
  {
    my ($firstname,$lastname);
    $lastname=$username;
    if ( $username =~ /^([^ ]*) (.*)$/ )
    {
      $firstname=$1;
      $lastname=$2;
    }
    print "givenName: ".$firstname."\n" if $firstname;
    if ( $username )
    {
      print "cn: ".$username."\n";
    }
    else
    {
      print "cn: NOT SET\n";
    }
    if ( $lastname )
    {
      print "sn: ".$lastname."\n";
    }
    else
    {
      print "sn: NOT SET\n";
    }
    print "gecos: ".$username."\n" if $username;
  }
  print "homeDirectory: ".$homebase."/".lc($user_name)."\n";
  print "loginShell: ".$loginshell."\n";
  print "shadowWarning: 7\n";
  print "shadowInactive: 2\n";
  print "smbHome: ".$smbhome."\n";
  print "homeDrive: ".$homedrive."\n";
  print "rid: ".$uid."\n";
  print "trustModel: fullaccess\n";
}

sub get_uid
{
  my $username=shift;
  my ($login,$pass,$uid,$gid) = getpwnam($username) or return $user_id_startat++;
  return $uid;
}

sub dump_group
{
  my ($group_name,$group_id,$group_flags,$group_secbyte,$group_props) = @_;
  my $gid=$group_id_startat++;
  print "dn: cn=".lc($group_name).",".$group_basedn."\n";
  print "objectClass: top\n";
  print "objectClass: posixGroup\n";
  print "objectClass: groupOfNames\n";
  print "cn: ".lc($group_name)."\n";
  print "gidNumber: ".$gid."\n";
  foreach (get_property_value($group_name,'0002','GROUP_MEMBERS'))
  {
    my ($user_name,$user_id);
    if (  /^(.*) ([0-9A-F]{8}) ([0-9]{4})$/ )
    {
      $user_name=$1;
      $user_id=$2;
      die if $3 ne '0001';
    }
    else
    {
      die;
    }
    print "memberUid: ".lc($user_name)."\n";
    print "member: uid=".lc($user_name).",".$user_basedn."\n";
  }
  my ($groupdesc,@dummy)=ansi2utf(get_property_value($group_name,'0002','IDENTIFICATION'));
  print "description: ".$groupdesc."\n" if $groupdesc;
}

sub get_gid
{
  my $group=shift;
  my ($name,$passwd,$gid,$members) = getgrnam($groupname) or return $group_id_startat++;
  return $gid;
}

# convert an ANSI string to a UTF-8 string
sub ansi2utf
{
  my $s=shift;
  my $res='';
  for (split(//,$s))
  {
    my $c=ord($_);
    if ($c<=127)
    {
      $res.=chr($c);
    }
    else
    {
      $res.=chr((($c>>6)&0x03)|0xc0).chr(($c&0x3f)|0x80);
    }
  }
  return $res;
}
